#include <msp/gl/matrix.h>
#include <msp/gl/meshbuilder.h>
#include <msp/gl/texture.h>
#include <msp/input/keys.h>
#include "entry.h"
#include "graphic.h"
#include "part.h"
#include "style.h"
#include "vslider.h"

using namespace std;

namespace Msp {
namespace GLtk {

Entry::Entry(const string &t):
	text(),
	multiline(false),
	edit_width(10),
	edit_height(1),
	edit_pos(0),
	first_row(0),
	visible_rows(1),
	text_part(0),
	slider(0),
	got_key_press(false)
{
	set_text(t);
}

void Entry::autosize_special(const Part &part, Geometry &ageom) const
{
	if(part.get_name()=="text")
	{
		const Sides &margin = part.get_margin();
		const GL::Font &font = style->get_font();
		unsigned en_width = static_cast<unsigned>(font.get_string_width("n")*style->get_font_size());
		ageom.w = max(ageom.w, edit_width*en_width+margin.left+margin.right);

		unsigned line_height = static_cast<unsigned>((font.get_ascent()-font.get_descent())*style->get_font_size());
		if(multiline)
		{
			unsigned line_spacing = style->get_font_size()*6/5;
			ageom.h = max(ageom.h, line_height+line_spacing*(edit_height-1)+margin.top+margin.bottom);
		}
		else
			ageom.h = max(ageom.h, line_height+margin.top+margin.bottom);
	}
	else if(part.get_name()=="slider" && multiline)
	{
		Geometry sgeom = part.get_geometry();
		if(!sgeom.w || !sgeom.h)
		{
			Geometry wgeom;
			slider->autosize(wgeom);
			if(!sgeom.w)
				sgeom.w = wgeom.w;
			if(!sgeom.h)
				sgeom.h = wgeom.h;
		}

		const Sides &margin = part.get_margin();
		ageom.w = max(ageom.w, sgeom.w+margin.left+margin.right);
		ageom.h = max(ageom.h, sgeom.h+margin.top+margin.bottom);
	}
}

void Entry::set_text(const string &t)
{
	text = t;
	edit_pos = text.size();

	if(multiline)
		check_view_range();

	rebuild();
}

void Entry::set_edit_size(unsigned w, unsigned h)
{
	edit_width = w;
	edit_height = h;
	signal_autosize_changed.emit();
}

void Entry::set_multiline(bool m)
{
	multiline = m;
	if(multiline)
	{
		if(!slider)
		{
			slider = new VSlider;
			add(*slider);
			slider->set_step(1);
			slider->signal_value_changed.connect(sigc::mem_fun(this, &Entry::slider_value_changed));
			reposition_slider();
		}
		check_view_range();
	}
}

void Entry::rebuild_special(const Part &part)
{
	if(part.get_name()=="text")
		text.build(part, state, geom, first_row, part_cache);
	else if(part.get_name()=="cursor")
	{
		const Graphic *graphic = part.get_graphic(state);
		if(!text_part || !graphic || !graphic->get_texture())
			return;

		unsigned row, col;
		text.offset_to_coords(edit_pos, row, col);

		if(row<first_row || row>=first_row+visible_rows)
			return;

		Geometry rgeom = text.coords_to_geometry(*text_part, geom, first_row, row, col);

		GL::MeshBuilder bld(part_cache.create_mesh(part, *graphic->get_texture()));
		bld.matrix() *= GL::Matrix::translation(rgeom.x, rgeom.y, 0);
		graphic->build(part.get_geometry().w, part.get_geometry().h, bld);
	}
	else
		Widget::rebuild_special(part);
}

void Entry::render_special(const Part &part, GL::Renderer &renderer) const
{
	if(part.get_name()=="slider" && multiline)
		slider->render(renderer);
}

void Entry::key_press(unsigned key, unsigned)
{
	got_key_press = true;
	if(key==Input::KEY_LEFT)
	{
		if(edit_pos>0)
			set_edit_position(edit_pos-1);
	}
	else if(key==Input::KEY_RIGHT)
	{
		if(edit_pos<text.size())
			set_edit_position(edit_pos+1);
	}
	else if(key==Input::KEY_DOWN && multiline)
	{
		unsigned row, col;
		text.offset_to_coords(edit_pos, row, col);
		set_edit_position(text.coords_to_offset(row+1, col));
	}
	else if(key==Input::KEY_UP && multiline)
	{
		unsigned row, col;
		text.offset_to_coords(edit_pos, row, col);
		set_edit_position(row>0 ? text.coords_to_offset(row-1, col) : 0);
	}
	else if(key==Input::KEY_BACKSPACE)
	{
		if(edit_pos>0)
		{
			text.erase(--edit_pos, 1);
			check_view_range();
			rebuild();
		}
	}
	else if(key==Input::KEY_ENTER)
	{
		if(multiline)
		{
			text.insert(edit_pos++, "\n");
			check_view_range();
			rebuild();
		}
		else
			signal_enter.emit();
	}
}

void Entry::character(wchar_t ch)
{
	if(got_key_press && ch>=' ')
	{
		text.insert(edit_pos, StringCodec::encode<StringCodec::Utf8>(StringCodec::ustring(1, ch)));
		++edit_pos;
		rebuild();
	}
}

void Entry::focus_out()
{
	Widget::focus_out();
	got_key_press = false;
}

void Entry::on_geometry_change()
{
	reposition_slider();

	if(multiline)
		check_view_range();
}

void Entry::on_style_change()
{
	text.set_style(style);

	if(!style)
	{
		text_part = 0;
		return;
	}

	text_part = style->get_part("text");

	reposition_slider();

	if(multiline)
		check_view_range();
}

void Entry::set_edit_position(unsigned ep)
{
	edit_pos = ep;
	check_view_range();
	rebuild();
}

void Entry::reposition_slider()
{
	if(!style || !slider)
		return;

	if(const Part *slider_part = style->get_part("slider"))
	{
		Geometry sgeom = slider_part->get_geometry();
		if(!sgeom.w || !sgeom.h)
		{
			Geometry wgeom;
			slider->autosize(wgeom);
			if(!sgeom.w)
				sgeom.w = wgeom.w;
			if(!sgeom.h)
				sgeom.h = wgeom.h;
		}

		slider_part->get_alignment().apply(sgeom, geom, slider_part->get_margin());
		slider->set_geometry(sgeom);
	}
}

void Entry::check_view_range()
{
	if(!multiline || !text_part)
		return;

	float font_size = style->get_font_size();
	unsigned line_spacing = static_cast<unsigned>(font_size*6/5);

	const Sides &margin = text_part->get_margin();
	visible_rows = max((geom.h-margin.top-margin.bottom)/line_spacing, 1U);

	unsigned row, col;
	text.offset_to_coords(edit_pos, row, col);

	if(first_row>row)
		first_row = row;
	else if(row>=first_row+visible_rows)
		first_row = row+1-visible_rows;

	unsigned scroll = max(text.get_n_lines(), visible_rows)-visible_rows;
	slider->set_range(0, scroll);
	slider->set_value(scroll-first_row);
}

void Entry::slider_value_changed(double value)
{
	if(text.get_n_lines()>visible_rows)
		first_row = text.get_n_lines()-visible_rows-static_cast<unsigned>(value);
}


Entry::Loader::Loader(Entry &e):
	DataFile::DerivedObjectLoader<Entry, Widget::Loader>(e)
{
	add("edit_size", &Entry::edit_width, &Entry::edit_height);
	add("multiline", &Loader::multiline);
}

void Entry::Loader::multiline(bool m)
{
	obj.set_multiline(m);
}

} // namespace GLtk
} // namespace Msp
