#ifndef MSP_GLTK_LIST_H_
#define MSP_GLTK_LIST_H_

#include <stdexcept>
#include <typeinfo>
#include <sigc++/signal.h>
#include "container.h"
#include "label.h"
#include "listdata.h"
#include "vslider.h"

namespace Msp {
namespace GLtk {

/**
Thrown if a list's item type is incompatible with its data.
*/
class incompatible_data: public std::logic_error
{
public:
	incompatible_data(const std::type_info &);
	virtual ~incompatible_data() throw() { }
};


/**
Shows a list of items, allowing the user to select one.  A slider is included
to allow scrolling through a long list.
*/
class List: virtual public Widget, private Container
{
public:
	class Loader: public DataFile::DerivedObjectLoader<List, Widget::Loader>
	{
	public:
		Loader(List &);
	private:
		void item(const std::string &);
	};

private:
	/// This exists to make disconnecting signals easier
	class DataObserver: public sigc::trackable
	{
	private:
		List &list;

	public:
		DataObserver(List &);

		void item_added(unsigned);
		void item_removed(unsigned);
		void cleared();
		void refresh_item(unsigned);
	};

public:
	class Item: public Container
	{
	public:
		virtual const char *get_class() const { return "listitem"; }

	protected:
		virtual void autosize_special(const Part &, Geometry &) const;
	public:
		void set_active(bool);

		virtual void render_special(const Part &, GL::Renderer &) const;
	};

private:
	class BasicItem: public Item
	{
	private:
		Label label;

	public:
		BasicItem(const std::string &);

	private:
		virtual void on_style_change();
	};

	class ItemFactory
	{
	protected:
		ItemFactory() { }
	public:
		virtual ~ItemFactory() { }

		virtual void set_data(const ListData &) = 0;
		virtual Item *create_item(unsigned) const = 0;
	};

	template<typename I>
	class TypedItemFactory: public ItemFactory
	{
	private:
		typedef typename I::ValueType ValueType;

		const ListDataStore<ValueType> *data;

	public:
		TypedItemFactory(const ListData &d)
		{ set_data(d); }

		virtual void set_data(const ListData &d)
		{
			if(const ListDataStore<ValueType> *ds = dynamic_cast<const ListDataStore<ValueType> *>(&d))
				data = ds;
			else
				throw incompatible_data(typeid(ValueType));
		}

		virtual Item *create_item(unsigned i) const
		{
			return new I(data->get(i));
		}
	};

public:
	sigc::signal<void, unsigned> signal_item_selected;

private:
	ListData *data;
	bool own_data;
	DataObserver *observer;
	ItemFactory *item_factory;
	int sel_index;
	unsigned first;
	unsigned max_scroll;
	unsigned view_size;

	VSlider slider;
	std::vector<Item *> items;

public:
	List();
	List(ListData &);
private:
	void init();
public:
	virtual ~List();

	virtual const char *get_class() const { return "list"; }

private:
	virtual void autosize_special(const Part &, Geometry &) const;

public:
	void set_data(ListData &);
	ListData &get_data() { return *data; }
	const ListData &get_data() const { return *data; }
private:
	void items_changed();

public:
	template<typename I>
	void set_item_type()
	{
		ItemFactory *f = new TypedItemFactory<I>(*data);
		delete item_factory;
		item_factory = f;
	}
private:
	Item *create_item(unsigned);

public:
	void set_view_size(unsigned);
	void set_view_all();

	void set_selected_index(int);
	int get_selected_index() const { return sel_index; }

private:
	virtual void render_special(const Part &, GL::Renderer &) const;

public:
	virtual void button_press(int, int, unsigned);
private:
	virtual void on_geometry_change();
	virtual void on_style_change();

	void reposition_slider();
	void item_autosize_changed();
	void reposition_items();
	void check_view_range();
	void slider_value_changed(double);
};

} // namespace GLtk
} // namespace Msp

#endif
