#include <algorithm>
#include <msp/core/maputils.h>
#include <msp/core/refptr.h>
#include "button.h"
#include "column.h"
#include "draghandle.h"
#include "dropdown.h"
#include "entry.h"
#include "grid.h"
#include "hslider.h"
#include "image.h"
#include "indicator.h"
#include "label.h"
#include "list.h"
#include "panel.h"
#include "part.h"
#include "row.h"
#include "stack.h"
#include "toggle.h"
#include "vslider.h"

using namespace std;

namespace Msp {
namespace GLtk {

Panel::Panel():
	layout(0)
{ }

Panel::~Panel()
{
	delete layout;
	layout = 0;
}

void Panel::set_layout(Layout *l)
{
	l->set_container(*this);
	delete layout;
	layout = l;
}

void Panel::autosize_special(const Part &part, Geometry &ageom) const
{
	if(part.get_name()=="children" && layout)
		layout->autosize(ageom);
}

void Panel::render_special(const Part &part, GL::Renderer &renderer) const
{
	if(part.get_name()=="children")
	{
		for(list<Container::Child *>::const_iterator i=children.begin(); i!=children.end(); ++i)
			if((*i)->widget->is_visible())
				(*i)->widget->render(renderer);
	}
}

void Panel::on_geometry_change()
{
	if(layout)
		layout->update();
}

void Panel::on_child_added(Widget &wdg)
{
	if(layout)
	{
		layout->add_widget(wdg);
		signal_autosize_changed.emit();
	}
}

void Panel::on_child_removed(Widget &wdg)
{
	if(layout)
	{
		layout->remove_widget(wdg);
		signal_autosize_changed.emit();
	}
}


Panel::Loader::Loader(Panel &p, map<string, Widget *> &m):
	DataFile::DerivedObjectLoader<Panel, Widget::Loader>(p),
	wdg_map(m),
	last_widget(0)
{
	add("button",    &Loader::child<Button>);
	add("column",    &Loader::arrangement<Column>);
	add("constraint",&Loader::constraint);
	add("draghandle",&Loader::child<DragHandle>);
	add("dropdown",  &Loader::child<Dropdown>);
	add("entry",     &Loader::child<Entry>);
	add("expand",    &Loader::expand);
	add("ghost",     &Loader::ghost);
	add("gravity",   &Loader::gravity);
	add("grid",      &Loader::grid);
	add("hslider",   &Loader::child<HSlider>);
	add("image",     &Loader::child<Image>);
	add("indicator", &Loader::child<Indicator>);
	add("label",     &Loader::child<Label>);
	add("layout",    &Loader::layout);
	add("list",      &Loader::child<List>);
	add("panel",     &Loader::panel);
	add("row",       &Loader::arrangement<Row>);
	add("stack",     &Loader::arrangement<Stack>);
	add("toggle",    &Loader::child<Toggle>);
	add("vslider",   &Loader::child<VSlider>);
}

Layout &Panel::Loader::get_layout()
{
	if(!obj.layout)
		obj.set_layout(new Layout);

	return *obj.layout;
}

Widget &Panel::Loader::get_last_widget()
{
	if(!last_widget)
		throw logic_error("no widget loaded");

	return *last_widget;
}

template<typename T>
void Panel::Loader::arrangement()
{
	T arr(get_layout());
	ArrangedLoader<T> ldr(*this, arr);
	load_sub_with(ldr);
}

template<typename T>
void Panel::Loader::child(const string &n)
{
	RefPtr<T> chl = new T();
	load_sub(*chl);
	obj.add(*chl.get());
	last_widget = wdg_map[n] = chl.release();
}

void Panel::Loader::constraint(Layout::ConstraintType type, const string &n)
{
	Widget &src = get_last_widget();
	Widget &tgt = *get_item(wdg_map, n);
	get_layout().add_constraint(src, type, tgt);
}

void Panel::Loader::expand(bool h, bool v)
{
	get_layout().set_expand(get_last_widget(), h, v);
}

void Panel::Loader::ghost(bool g)
{
	get_layout().set_ghost(get_last_widget(), g);
}

void Panel::Loader::gravity(int h, int v)
{
	get_layout().set_gravity(get_last_widget(), h, v);
}

void Panel::Loader::grid(unsigned cols)
{
	Grid grd(get_layout(), cols);
	ArrangedLoader<Grid> ldr(*this, grd);
	load_sub_with(ldr);
}

void Panel::Loader::layout()
{
	Layout::Loader ldr(get_layout(), wdg_map);
	load_sub_with(ldr);
}

void Panel::Loader::panel(const string &n)
{
	RefPtr<Panel> p = new Panel();
	load_sub(*p, wdg_map);
	obj.add(*p.get());
	last_widget = wdg_map[n] = p.release();
}


template<typename T>
Panel::ArrangedLoader<T>::ArrangedLoader(Loader &ldr, T &arr):
	arr_loader(arr)
{
	add_auxiliary_loader(ldr);
	add_auxiliary_loader(arr_loader);
}

} // namespace GLtk
} // namespace Msp
