#ifndef MSP_GLTK_WIDGET_H_
#define MSP_GLTK_WIDGET_H_

#include <string>
#include <msp/datafile/objectloader.h>
#include <msp/gl/renderer.h>
#include "geometry.h"
#include "partcache.h"
#include "state.h"

namespace Msp {
namespace GLtk {

class Container;
class Part;
class Resources;
class Style;

/**
Base class for all widgets.  Derived classes should call update_style in 
constructor, because it can't be done correctly in the Widget constructor.
*/
class Widget
{
	friend class Container;

public:
	class Loader: public DataFile::ObjectLoader<Widget>
	{
	public:
		Loader(Widget &);
	private:
		void position(int, int);
		void size(unsigned, unsigned);
		void style(const std::string &);
	};

	sigc::signal<void, bool> signal_visibility_changed;
	sigc::signal<void> signal_autosize_changed;
	sigc::signal<void> signal_request_focus;
	sigc::signal<void> signal_grab_pointer;
	sigc::signal<void> signal_ungrab_pointer;

protected:
	Geometry geom;
	std::string style_name;
	const Style *style;
	State state;
	bool visible;
	bool focusable;
	Container *parent;
	std::string tooltip;
	PartCache part_cache;

	Widget();
private:
	Widget(const Widget &);
	Widget &operator=(const Widget &);
public:
	virtual ~Widget();

	/// Returns the name of the widget class.  Used for style lookup.
	virtual const char *get_class() const { return "widget"; }

	void set_position(int, int);
	void set_size(unsigned, unsigned);
	void autosize();
	void autosize(Geometry &) const;
protected:
	virtual void autosize_special(const Part &, Geometry &) const { };
public:
	void set_geometry(const Geometry &);
	const Geometry &get_geometry() const { return geom; }

	void map_coords_to_ancestor(int &, int &, const Widget &) const;

protected:
	/** Sets the widget's parent Container.  The widget must be unparented when
	calling this function with a non-null parameter. */
	void set_parent(Container *);
public:
	Container *get_parent() const { return parent; }

	/** Finds the closest ancestor of a specific type. */
	template<typename T>
	T *find_ancestor() const
	{
		for(Widget *w=parent; w; w=w->get_parent())
			if(T *tw = dynamic_cast<T *>(w))
				return tw;
		return 0;
	}

	/** Sets the widget style.  The name of the resource to be looked up is
	constructed by concatenating the widget class and the style name with a
	dash. */
	void set_style(const std::string &);
	const Style &get_style() const { return *style; }

protected:
	/** Gets a style object from the resource collection based on the class and
	style names of the widget. */
	void update_style();

public:
	void set_tooltip(const std::string &);
	const std::string &get_tooltip() const { return tooltip; }

	void set_visible(bool);
	bool is_visible() const { return visible; }
	void set_focusable(bool);
	bool is_focusable() const { return focusable; }
	void set_focus();
	void set_enabled(bool);
	bool is_enabled() const { return !(state&DISABLED); }

protected:
	void set_state(State s) { set_state(s, s); }
	void clear_state(State s) { set_state(s, NORMAL); }
	void set_state(State, State);
public:
	State get_state() const { return state; }

protected:
	void rebuild();
	virtual void rebuild_special(const Part &);

public:
	void render(GL::Renderer &) const;
protected:
	virtual void render_special(const Part &, GL::Renderer &) const { }

public:
	// Events
	virtual void button_press(int, int, unsigned) { }
	virtual void button_release(int, int, unsigned) { }
	virtual void pointer_motion(int, int) { }
	virtual void pointer_enter();
	virtual void pointer_leave();
	virtual void key_press(unsigned, unsigned) { }
	virtual void key_release(unsigned, unsigned) { }
	virtual void character(wchar_t) { }
	virtual void focus_in();
	virtual void focus_out();
protected:
	virtual void on_geometry_change() { }
	virtual void on_style_change() { }
	virtual void on_reparent() { }
};

} // namespace GLtk
} // namespace Msp

#endif
